/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotAndObjectToIndex
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ObjectInstanceToIndex.h"

#include <SimoxUtility/shapes/AxisAlignedBoundingBox.h>
#include <SimoxUtility/shapes/OrientedBox.h>

#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem_index/aron/Named.aron.generated.h>
#include <RobotAPI/libraries/armem_index/aron/Spatial.aron.generated.h>
#include <RobotAPI/libraries/armem_objects/memory_ids.h>
#include <RobotAPI/libraries/armem_objects/utils.h>
#include <RobotAPI/libraries/aron/common/aron_conversions/simox.h>
#include <RobotAPI/libraries/aron/common/aron_conversions/stl.h>

namespace armarx::armem::objects
{

    void
    ObjectInstanceToIndex::fetchAndCommitObjectInstances(
        const std::vector<armem::MemoryID>& updatedObjectInstanceSnapshotIDs)
    {
        ARMARX_CHECK(objectPoseClient.isConnected());

        // Fetch the latest poses.
        const objpose::ObjectPoseSeq objectPoses = objectPoseClient.fetchObjectPoses();

        const auto filtered = filterObjectPoses(objectPoses, updatedObjectInstanceSnapshotIDs);


        // Prepare the commit.

        armem::Commit commit;

        for (const objpose::ObjectPose* objectPosePtr : filtered)
        {
            const objpose::ObjectPose& objectPose = *objectPosePtr;

            armem::MemoryID objectInstanceID =
                armem::objects::reconstructObjectInstanceID(objectPose);

            // Spatial

            std::optional<simox::OrientedBoxf> oobb = objectPose.oobbGlobal();
            if (oobb.has_value())
            {
                simox::AxisAlignedBoundingBox aabb =
                    simox::AxisAlignedBoundingBox::from_points(oobb->corners());

                armem::index::arondto::Spatial spatial;
                armem::toAron(spatial.id, objectInstanceID);
                aron::toAron(spatial.oobbGlobal, oobb);
                toAron(spatial.aabbGlobal, aabb);

                armem::EntityUpdate& update = commit.add();
                update.entityID = indexSpatialProviderSegmentID.withEntityName(
                    objectInstanceID.getEntityID().str());
                update.referencedTime = objectPose.timestamp;
                update.instancesData = {spatial.toAron()};
            }

            // Named

            // Load object class information.
            std::optional<ObjectInfo> info =
                objectPoseClient.getObjectFinder().findObject(objectPose.objectID);
            if (info.has_value())
            {
                std::optional<std::vector<std::string>> recognized, spoken;
                recognized = info->loadRecognizedNames();
                spoken = info->loadSpokenNames();

                armem::index::arondto::Named named;
                armem::toAron(named.id, objectInstanceID);

                if (recognized.has_value())
                {
                    named.names.recognized = recognized.value();
                }
                else
                {
                    named.names.recognized = {info->className()};
                }

                if (spoken.has_value())
                {
                    named.names.spoken = spoken.value();
                }
                else
                {
                    named.names.spoken = {info->className()};
                }

                armem::EntityUpdate& update = commit.add();
                update.entityID = indexNamedProviderSegmentID.withEntityName(
                    objectInstanceID.getEntityID().str());
                update.referencedTime = objectPose.timestamp;
                update.instancesData = {named.toAron()};
            }
        }

        // Commit.
        indexSpatialMemoryWriter.commit(commit);
    }

    std::vector<const objpose::ObjectPose*>
    ObjectInstanceToIndex::filterObjectPoses(const objpose::ObjectPoseSeq& objectPoses,
                                             const std::vector<MemoryID>& updatedSnapshotIDs)
    {
        // Returns true to keep the item, false to skip it.
        auto filter = [this, &updatedSnapshotIDs](const objpose::ObjectPose& objectPose)
        {
            auto it = state.latestUpdateDateTimes.find(objectPose.objectID);
            if (it == state.latestUpdateDateTimes.end())
            {
                // Never encountered that before, commit it.
                return true;
            }

            const armarx::DateTime& latestTime = it->second;
            armarx::DateTime nextDueTime = latestTime + params.maxFrequency.toCycleDuration();
            if (objectPose.timestamp < nextDueTime)
            {
                // Skip.
                return false;
            }

            armem::MemoryID objectInstanceID =
                armem::objects::reconstructObjectInstanceID(objectPose);
            bool found = false;
            for (const MemoryID& updatedSnapshotID : updatedSnapshotIDs)
            {
                if (armem::contains(updatedSnapshotID, objectInstanceID))
                {
                    found = true;
                    break;
                }
            }
            return found;
        };


        std::vector<const objpose::ObjectPose*> filtered;
        for (const objpose::ObjectPose& objectPose : objectPoses)
        {
            if (filter(objectPose))
            {
                filtered.push_back(&objectPose);
                state.latestUpdateDateTimes[objectPose.objectID] = objectPose.timestamp;
            }
        }

        return filtered;
    }

} // namespace armarx::armem::objects
