/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::aron_cpp_to_python_conv_test
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/interface/aron/test/AronConversionTestInterface.h>

namespace armarx::aron::test
{

    /**
     * @brief Helper class for implementing distributed ARON conversion tests
     * based on the `dti::AronConversionTestInterfacePrx`.
     *
     * Example usage:
     *
     * @code
     *
     * armarx::aron::test::dti::AronConversionTestInterfacePrx pythonComponent = ...;
     *
     * auto myAronClassProbeFn = []()
     * {
     *     my::arondto::MyAronClass probe;
     *     probe.data = "42";
     *     return probe,
     * };
     *
     * armarx::aron::test::AronConversionTester tester(pythonComponent);
     *
     * tester.test<my::arondto::MyAronClass>(myAronClassProbeFn, "MyAronClass");
     *
     * @endcode
     *
     * Note that the `pythonComponent` must point to a corresponding
     * implementation of the `AronConversionTestInterfacePrx`.
     */
    class AronConversionTester
    {
    public:
        template <class AronClassT>
        using ProbeFn = std::function<AronClassT()>;


    public:
        AronConversionTester(dti::AronConversionTestInterfacePrx interface);

        /**
         * @brief Test the conversion of a specific ARON class.
         *
         * @param probeFn A factory function creating a test instance of the
         *      ARON class.
         * @param aronClassName The name of the ARON class. Can be used by the
         *      other component to decide which class to convert to.
         */
        template <class AronClassT>
        void
        test(ProbeFn<AronClassT> probeFn, const std::string& aronClassName)
        {
            std::stringstream ss;
            ss << "Test for ARON class '" << aronClassName << "': ";

            const AronClassT probe = probeFn();

            dto::TestAronConversionRequest req;
            req.aronClassName = aronClassName;
            req.probe = probe.toAronDTO();

            dto::TestAronConversionResponse res = interface->testAronConversion(req);

            if (res.success)
            {
                const AronClassT probeOut = AronClassT::FromAron(res.probe);
                ARMARX_CHECK(probeOut == probe);

                ARMARX_IMPORTANT << ss.str() << "Success";
            }
            else
            {
                ARMARX_WARNING << ss.str() << "Conversion in Python component failed: \n"
                               << res.errorMessage;
            }
        }

    public:
        dti::AronConversionTestInterfacePrx interface;
    };

} // namespace armarx::aron::test
