#pragma once

#include <iostream>
#include <mutex>

#include <QLayout>
#include <QLayoutItem>
#include <QSpinBox>
#include <QSplitter>
#include <QTreeWidgetItem>
#include <QWidget>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

namespace armarx::gui
{
    /**
     * @brief Clear a layout.
     *
     * Recursively take and delete all items in `layout`.
     *
     * @param layout The layout.
     */
    void clearLayout(QLayout* layout);

    void clearSplitter(QSplitter* splitter);

    /**
     * @brief Clear a tree widget item
     *
     * Take and delete all children of `item`.
     *
     * @param item The tree widget item.
     */
    void clearItem(QTreeWidgetItem* item);

    template <class WidgetT>
    void
    replaceWidget(WidgetT*& old, QWidget* neu, QLayout* parentLayout)
    {
        ARMARX_CHECK(old);
        ARMARX_CHECK(neu);
        ARMARX_CHECK(parentLayout);
        QLayoutItem* oldItem = parentLayout->replaceWidget(old, neu);
        if (oldItem)
        {
            delete oldItem;
            delete old;
            old = nullptr;
        }
    }

    template <class WidgetT>
    void
    replaceWidget(WidgetT*& old, QWidget* neu, QSplitter*& parentLayout)
    {
        ARMARX_CHECK(old);
        ARMARX_CHECK(neu);
        ARMARX_CHECK(parentLayout);
        int index = parentLayout->indexOf(old);
        ARMARX_CHECK(index >= 0) << "Could not find widget in splitter";
        QWidget* oldItem = parentLayout->replaceWidget(index, neu);
        if (oldItem)
        {
            // deleting the widget(s) directly will segfault for some reason
            oldItem->deleteLater();
            old->deleteLater();
            old = nullptr;
        }
    }

    /**
     * @brief Let items in `layout` be children of a splitter.
     *
     * Items in `layout` are moved to a new `QSplitter`, which will
     * be the only child of `layout`.
     * If `layout` is a Q{H,V}BoxLayout, the respective orientation
     * will be adopted.
     *
     * @param The parent layout.
     * @return The splitter item.
     */
    QSplitter* useSplitter(QLayout* layout);

    // Source: https://stackoverflow.com/a/26538572
    class LeadingZeroSpinBox : public QSpinBox
    {
        using QSpinBox::QSpinBox;

    public:
        LeadingZeroSpinBox(int numDigits, int base);

        QString textFromValue(int value) const override;

    private:
        int numDigits;
        int base;
    };
} // namespace armarx::gui
